@@ -1,9 +1,9 @@
 /*
  * security/ccsecurity/gc.c
  *
- * Copyright (C) 2005-2009  NTT DATA CORPORATION
+ * Copyright (C) 2005-2011  NTT DATA CORPORATION
  *
- * Version: 1.7.1   2009/11/11
+ * Version: 1.7.3   2011/04/01
  *
  * This file is applicable to both 2.4.30 and 2.6.11 and later.
  * See README.ccs for ChangeLog.
@@ -16,6 +16,7 @@
 #include <linux/kthread.h>
 #endif
 
+/* Index numbers for garbage collection. */
 enum ccs_gc_id {
 	CCS_ID_RESERVEDPORT,
 	CCS_ID_ADDRESS_GROUP,
@@ -39,74 +40,166 @@ enum ccs_gc_id {
 	CCS_ID_DOMAIN
 };
 
+/* Structure for garbage collection. */
 struct ccs_gc_entry {
 	struct list_head list;
-	int type;
-	void *element;
+	int type; /* = one of values in "enum ccs_gc_id" */
+	struct list_head *element;
 };
-static LIST_HEAD(ccs_gc_queue);
-static DEFINE_MUTEX(ccs_gc_mutex);
+/* List of entries to be deleted. */
+static LIST_HEAD(ccs_gc_list);
+/* Length of list. */
+static int ccs_gc_list_len;
 
-/* Caller holds ccs_policy_lock mutex. */
-static bool ccs_add_to_gc(const int type, void *element)
+/**
+ * ccs_add_to_gc - Add an entry to to be deleted list.
+ *
+ * @type:    Type of this entry.
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns true on success, false otherwise.
+ *
+ * Caller holds ccs_policy_lock mutex.
+ *
+ * Adding an entry needs kmalloc(). Thus, if we try to add thousands of
+ * entries at once, it will take too long time. Thus, do not add more than 128
+ * entries per a scan. But to be able to handle worst case where all entries
+ * are in-use, we accept one more entry per a scan.
+ *
+ * If we use singly linked list using "struct list_head"->prev (which is
+ * LIST_POISON2), we can avoid kmalloc().
+ */
+static bool ccs_add_to_gc(const int type, struct list_head *element)
 {
-	struct ccs_gc_entry *entry = kzalloc(sizeof(*entry), GFP_ATOMIC);
+	struct ccs_gc_entry *entry = kzalloc(sizeof(*entry), CCS_GFP_FLAGS);
 	if (!entry)
 		return false;
 	entry->type = type;
 	entry->element = element;
-	list_add(&entry->list, &ccs_gc_queue);
-	return true;
+	list_add(&entry->list, &ccs_gc_list);
+	list_del_rcu(element);
+	return ccs_gc_list_len++ < 128;
 }
 
-static size_t ccs_del_allow_read(struct ccs_globally_readable_file_entry *ptr)
+/**
+ * ccs_del_allow_read - Delete members in "struct ccs_globally_readable_file_entry".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_allow_read(struct list_head *element)
 {
+	struct ccs_globally_readable_file_entry *ptr =
+		container_of(element, typeof(*ptr), list);
 	ccs_put_name(ptr->filename);
 	return sizeof(*ptr);
 }
 
-static size_t ccs_del_allow_env(struct ccs_globally_usable_env_entry *ptr)
+/**
+ * ccs_del_allow_env - Delete members in "struct ccs_globally_usable_env_entry".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_allow_env(struct list_head *element)
 {
+	struct ccs_globally_usable_env_entry *ptr =
+		container_of(element, typeof(*ptr), list);
 	ccs_put_name(ptr->env);
 	return sizeof(*ptr);
 }
 
-static size_t ccs_del_file_pattern(struct ccs_pattern_entry *ptr)
+/**
+ * ccs_del_file_pattern - Delete members in "struct ccs_pattern_entry".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_file_pattern(struct list_head *element)
 {
+	struct ccs_pattern_entry *ptr =
+		container_of(element, typeof(*ptr), list);
 	ccs_put_name(ptr->pattern);
 	return sizeof(*ptr);
 }
 
-static size_t ccs_del_no_rewrite(struct ccs_no_rewrite_entry *ptr)
+/**
+ * ccs_del_no_rewrite - Delete members in "struct ccs_no_rewrite_entry".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_no_rewrite(struct list_head *element)
 {
+	struct ccs_no_rewrite_entry *ptr =
+		container_of(element, typeof(*ptr), list);
 	ccs_put_name(ptr->pattern);
 	return sizeof(*ptr);
 }
 
-static size_t ccs_del_domain_initializer(struct ccs_domain_initializer_entry *
-					 ptr)
+/**
+ * ccs_del_domain_initializer - Delete members in "struct ccs_domain_initializer_entry".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_domain_initializer(struct list_head *element)
 {
+	struct ccs_domain_initializer_entry *ptr =
+		container_of(element, typeof(*ptr), list);
 	ccs_put_name(ptr->domainname);
 	ccs_put_name(ptr->program);
 	return sizeof(*ptr);
 }
 
-static size_t ccs_del_domain_keeper(struct ccs_domain_keeper_entry *ptr)
+/**
+ * ccs_del_domain_keeper - Delete members in "struct ccs_domain_keeper_entry".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_domain_keeper(struct list_head *element)
 {
+	struct ccs_domain_keeper_entry *ptr =
+		container_of(element, typeof(*ptr), list);
 	ccs_put_name(ptr->domainname);
 	ccs_put_name(ptr->program);
 	return sizeof(*ptr);
 }
 
-static size_t ccs_del_aggregator(struct ccs_aggregator_entry *ptr)
+/**
+ * ccs_del_aggregator - Delete members in "struct ccs_aggregator_entry".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_aggregator(struct list_head *element)
 {
+	struct ccs_aggregator_entry *ptr =
+		container_of(element, typeof(*ptr), list);
 	ccs_put_name(ptr->original_name);
 	ccs_put_name(ptr->aggregated_name);
 	return sizeof(*ptr);
 }
 
-static size_t ccs_del_manager(struct ccs_policy_manager_entry *ptr)
+/**
+ * ccs_del_manager - Delete members in "struct ccs_policy_manager_entry".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_manager(struct list_head *element)
 {
+	struct ccs_policy_manager_entry *ptr =
+		container_of(element, typeof(*ptr), list);
 	ccs_put_name(ptr->manager);
 	return sizeof(*ptr);
 }
@@ -136,7 +229,7 @@ static bool ccs_used_by_task(struct ccs_domain_info *domain)
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
 	struct task_struct *g;
 	struct task_struct *t;
-	read_lock(&tasklist_lock);
+	ccs_tasklist_lock();
 	do_each_thread(g, t) {
 		if (!(t->ccs_flags & CCS_TASK_IS_IN_EXECVE)) {
 			smp_mb(); /* Avoid out of order execution. */
@@ -147,10 +240,10 @@ static bool ccs_used_by_task(struct ccs_domain_info *domain)
 		goto out;
 	} while_each_thread(g, t);
  out:
-	read_unlock(&tasklist_lock);
+	ccs_tasklist_unlock();
 #else
 	struct task_struct *p;
-	read_lock(&tasklist_lock);
+	ccs_tasklist_lock();
 	for_each_process(p) {
 		if (!(p->ccs_flags & CCS_TASK_IS_IN_EXECVE)) {
 			smp_mb(); /* Avoid out of order execution. */
@@ -160,14 +253,22 @@ static bool ccs_used_by_task(struct ccs_domain_info *domain)
 		in_use = true;
 		break;
 	}
-	read_unlock(&tasklist_lock);
+	ccs_tasklist_unlock();
 #endif
 	return in_use;
 }
 
-static size_t ccs_del_acl(struct ccs_acl_info *acl)
+/**
+ * ccs_del_acl - Delete members in "struct ccs_acl_info".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_acl(struct list_head *element)
 {
 	size_t size;
+	struct ccs_acl_info *acl = container_of(element, typeof(*acl), list);
 	ccs_put_condition(acl->cond);
 	switch (acl->type) {
 	case CCS_TYPE_PATH_ACL:
@@ -269,41 +370,92 @@ static size_t ccs_del_acl(struct ccs_acl_info *acl)
 		break;
 	default:
 		size = 0;
-		printk(KERN_WARNING "Unknown type\n");
 		break;
 	}
 	return size;
 }
 
-static size_t ccs_del_domain(struct ccs_domain_info *domain)
+/**
+ * ccs_del_domain - Delete members in "struct ccs_domain_info".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()) on success, 0 otherwise.
+ */
+static size_t ccs_del_domain(struct list_head *element)
 {
 	struct ccs_acl_info *acl;
 	struct ccs_acl_info *tmp;
+	struct ccs_domain_info *domain =
+		container_of(element, typeof(*domain), list);
+	/*
+	 * We need to recheck domain at this point.
+	 *
+	 * (1) Reader starts SRCU section upon execve().
+	 * (2) Reader traverses ccs_domain_list and finds this domain.
+	 * (3) Writer marks this domain as deleted.
+	 * (4) Garbage collector removes this domain from ccs_domain_list
+	 *     because this domain is marked as deleted and used by nobody.
+	 * (5) Reader saves reference to this domain into
+	 *     "struct task_struct"->ccs_domain_info .
+	 * (6) Reader finishes execve() operation and starts using this domain.
+	 * (7) Garbage collector waits for SRCU synchronization.
+	 * (8) Garbage collector kfree() this domain.
+	 *
+	 * By rechecking whether this domain is used by somebody or not at (8),
+	 * we can solve this race problem.
+	 */
 	if (ccs_used_by_task(domain))
 		return 0;
 	list_for_each_entry_safe(acl, tmp, &domain->acl_info_list, list) {
-		size_t size = ccs_del_acl(acl);
+		size_t size = ccs_del_acl(&acl->list);
 		ccs_memory_free(acl, size);
 	}
 	ccs_put_name(domain->domainname);
 	return sizeof(*domain);
 }
 
-static size_t ccs_del_path_group_member(struct ccs_path_group_member *member)
+/**
+ * ccs_del_path_group_member - Delete members in "struct ccs_path_group_member".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_path_group_member(struct list_head *element)
 {
+	struct ccs_path_group_member *member =
+		container_of(element, typeof(*member), list);
 	ccs_put_name(member->member_name);
 	return sizeof(*member);
 }
 
-static size_t ccs_del_path_group(struct ccs_path_group *group)
+/**
+ * ccs_del_path_group - Delete members in "struct ccs_path_group".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_path_group(struct list_head *element)
 {
+	struct ccs_path_group *group =
+		container_of(element, typeof(*group), list);
 	ccs_put_name(group->group_name);
 	return sizeof(*group);
 }
 
-static size_t ccs_del_address_group_member
-(struct ccs_address_group_member *member)
+/**
+ * ccs_del_address_group_member - Delete members in "struct ccs_address_group_member".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_address_group_member(struct list_head *element)
 {
+	struct ccs_address_group_member *member =
+		container_of(element, typeof(*member), list);
 	if (member->is_ipv6) {
 		ccs_put_ipv6_address(member->min.ipv6);
 		ccs_put_ipv6_address(member->max.ipv6);
@@ -311,40 +463,84 @@ static size_t ccs_del_address_group_member
 	return sizeof(*member);
 }
 
-static size_t ccs_del_address_group(struct ccs_address_group *group)
+/**
+ * ccs_del_address_group - Delete members in "struct ccs_address_group".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_address_group(struct list_head *element)
 {
+	struct ccs_address_group *group =
+		container_of(element, typeof(*group), list);
 	ccs_put_name(group->group_name);
 	return sizeof(*group);
 }
 
-static size_t ccs_del_number_group_member
-(struct ccs_number_group_member *member)
+/**
+ * ccs_del_number_group_member - Delete members in "struct ccs_number_group_member".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_number_group_member(struct list_head *element)
 {
+	struct ccs_number_group_member *member =
+		container_of(element, typeof(*member), list);
 	return sizeof(*member);
 }
 
-static size_t ccs_del_number_group(struct ccs_number_group *group)
+/**
+ * ccs_del_number_group - Delete members in "struct ccs_number_group".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_number_group(struct list_head *element)
 {
+	struct ccs_number_group *group =
+		container_of(element, typeof(*group), list);
 	ccs_put_name(group->group_name);
 	return sizeof(*group);
 }
 
-static size_t ccs_del_reservedport(struct ccs_reserved_entry *ptr)
-{
-	return sizeof(*ptr);
-}
-
-static size_t ccs_del_ipv6_address(struct ccs_ipv6addr_entry *ptr)
+/**
+ * ccs_del_reservedport - Delete members in "struct ccs_reserved_entry".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_reservedport(struct list_head *element)
 {
+	struct ccs_reserved_entry *ptr =
+		container_of(element, typeof(*ptr), list);
 	return sizeof(*ptr);
 }
 
 /**
- * ccs_del_conditiopn - Delete condition part.
+ * ccs_del_ipv6_address - Delete members in "struct ccs_ipv6addr_entry".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_ipv6_address(struct list_head *element)
+{
+	struct ccs_ipv6addr_entry *ptr =
+		container_of(element, typeof(*ptr), list);
+	return sizeof(*ptr);
+}
+
+/**
+ * ccs_del_condition - Delete members in "struct ccs_condition".
  *
  * @cond: Pointer to "struct ccs_condition".
  *
- * Returns size of condition in bytes.
+ * Returns size of @cond (for later kfree()).
  */
 size_t ccs_del_condition(struct ccs_condition *cond)
 {
@@ -377,106 +573,110 @@ size_t ccs_del_condition(struct ccs_condition *cond)
 	return cond->size;
 }
 
-static size_t ccs_del_name(const struct ccs_name_entry *ptr)
+/**
+ * ccs_del_name - Delete members in "struct ccs_name_entry".
+ *
+ * @element: Pointer to "struct list_head".
+ *
+ * Returns size of @element (for later kfree()).
+ */
+static size_t ccs_del_name(struct list_head *element)
 {
+	const struct ccs_name_entry *ptr =
+		container_of(element, typeof(*ptr), list);
 	return ptr->size;
 }
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 19)
+/* Lock for syscall users. */
+struct srcu_struct ccs_ss;
+#endif
+
 /*
- * 2.6.19 has SRCU support but it triggers general protection fault in my
- * environment. Thus, I use SRCU for 2.6.20 and later.
- */
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 20)
-
-/* Lock for GC. */
-static struct srcu_struct ccs_ss;
-
-/**
- * ccs_gc_init - Initialize garbage collector.
+ * Lock for /proc/ccs/ users.
  *
- * Returns 0.
+ * Currently, we hold SRCU lock upon open() and release upon close().
+ * Thus, kernel complains about returning to userspace with SRCU lock held.
+ * Therefore, non-SRCU lock is used for suppressing the kernel's complain
+ * messages. Modifying to hold/release SRCU lock upon each read()/write() is
+ * to-do list.
+ *
+ * Also used for syscall users for 2.6.18 and earlier kernels because
+ * they don't have SRCU support.
  */
-static int __init ccs_gc_init(void)
-{
-	if (init_srcu_struct(&ccs_ss))
-		panic("Out of memory.");
-	return 0;
-}
-core_initcall(ccs_gc_init);
-
-int ccs_read_lock(void)
-{
-	return srcu_read_lock(&ccs_ss);
-}
-
-void ccs_read_unlock(const int idx)
-{
-	srcu_read_unlock(&ccs_ss, idx);
-}
-
-static inline void ccs_synchronize_srcu(void)
-{
-	synchronize_srcu(&ccs_ss);
-}
-
-#else
-
-/* Lock for GC. */
 static struct {
 	int counter_idx;
 	int counter[2];
-} ccs_gc;
+} ccs_counter;
+/* Lock for protecting counter. */
 static DEFINE_SPINLOCK(ccs_counter_lock);
 
-int ccs_read_lock(void)
+/**
+ * ccs_lock - Hold non-SRCU lock.
+ *
+ * Returns index number which has to be passed to ccs_unlock().
+ */
+int ccs_lock(void)
 {
 	int idx;
 	spin_lock(&ccs_counter_lock);
-	idx = ccs_gc.counter_idx;
-	ccs_gc.counter[idx]++;
+	idx = ccs_counter.counter_idx;
+	ccs_counter.counter[idx]++;
 	spin_unlock(&ccs_counter_lock);
 	return idx;
 }
 
-void ccs_read_unlock(const int idx)
+/**
+ * ccs_unlock - Release non-SRCU lock.
+ *
+ * @idx: Index number returned by ccs_lock().
+ */
+void ccs_unlock(const int idx)
 {
 	spin_lock(&ccs_counter_lock);
-	ccs_gc.counter[idx]--;
+	ccs_counter.counter[idx]--;
 	spin_unlock(&ccs_counter_lock);
 }
 
-static void ccs_synchronize_srcu(void)
+/**
+ * ccs_synchronize_counter - Wait for SRCU grace period.
+ */
+static void ccs_synchronize_counter(void)
 {
 	int idx;
 	int v;
 	spin_lock(&ccs_counter_lock);
-	idx = ccs_gc.counter_idx;
-	ccs_gc.counter_idx ^= 1;
-	v = ccs_gc.counter[idx];
+	idx = ccs_counter.counter_idx;
+	ccs_counter.counter_idx ^= 1;
+	v = ccs_counter.counter[idx];
 	spin_unlock(&ccs_counter_lock);
 	while (v) {
 		ssleep(1);
 		spin_lock(&ccs_counter_lock);
-		v = ccs_gc.counter[idx];
+		v = ccs_counter.counter[idx];
 		spin_unlock(&ccs_counter_lock);
 	}
 }
 
-#endif
-
+/**
+ * ccs_collect_entry - Scan lists for deleted elements.
+ */
 static void ccs_collect_entry(void)
 {
-	mutex_lock(&ccs_policy_lock);
+	int i;
+	int idx;
+	if (mutex_lock_interruptible(&ccs_policy_lock))
+		return;
+	idx = ccs_read_lock();
 	{
 		struct ccs_globally_readable_file_entry *ptr;
 		list_for_each_entry_rcu(ptr, &ccs_globally_readable_list,
 					list) {
 			if (!ptr->is_deleted)
 				continue;
-			if (ccs_add_to_gc(CCS_ID_GLOBALLY_READABLE, ptr))
-				list_del_rcu(&ptr->list);
-			else
-				break;
+			if (!ccs_add_to_gc(CCS_ID_GLOBALLY_READABLE,
+					   &ptr->list))
+				goto unlock;
 		}
 	}
 	{
@@ -485,10 +685,8 @@ static void ccs_collect_entry(void)
 					list) {
 			if (!ptr->is_deleted)
 				continue;
-			if (ccs_add_to_gc(CCS_ID_GLOBAL_ENV, ptr))
-				list_del_rcu(&ptr->list);
-			else
-				break;
+			if (!ccs_add_to_gc(CCS_ID_GLOBAL_ENV, &ptr->list))
+				goto unlock;
 		}
 	}
 	{
@@ -496,10 +694,8 @@ static void ccs_collect_entry(void)
 		list_for_each_entry_rcu(ptr, &ccs_pattern_list, list) {
 			if (!ptr->is_deleted)
 				continue;
-			if (ccs_add_to_gc(CCS_ID_PATTERN, ptr))
-				list_del_rcu(&ptr->list);
-			else
-				break;
+			if (!ccs_add_to_gc(CCS_ID_PATTERN, &ptr->list))
+				goto unlock;
 		}
 	}
 	{
@@ -507,10 +703,8 @@ static void ccs_collect_entry(void)
 		list_for_each_entry_rcu(ptr, &ccs_no_rewrite_list, list) {
 			if (!ptr->is_deleted)
 				continue;
-			if (ccs_add_to_gc(CCS_ID_NO_REWRITE, ptr))
-				list_del_rcu(&ptr->list);
-			else
-				break;
+			if (!ccs_add_to_gc(CCS_ID_NO_REWRITE, &ptr->list))
+				goto unlock;
 		}
 	}
 	{
@@ -519,10 +713,9 @@ static void ccs_collect_entry(void)
 					list) {
 			if (!ptr->is_deleted)
 				continue;
-			if (ccs_add_to_gc(CCS_ID_DOMAIN_INITIALIZER, ptr))
-				list_del_rcu(&ptr->list);
-			else
-				break;
+			if (!ccs_add_to_gc(CCS_ID_DOMAIN_INITIALIZER,
+					   &ptr->list))
+				goto unlock;
 		}
 	}
 	{
@@ -530,10 +723,8 @@ static void ccs_collect_entry(void)
 		list_for_each_entry_rcu(ptr, &ccs_domain_keeper_list, list) {
 			if (!ptr->is_deleted)
 				continue;
-			if (ccs_add_to_gc(CCS_ID_DOMAIN_KEEPER, ptr))
-				list_del_rcu(&ptr->list);
-			else
-				break;
+			if (!ccs_add_to_gc(CCS_ID_DOMAIN_KEEPER, &ptr->list))
+				goto unlock;
 		}
 	}
 	{
@@ -541,10 +732,8 @@ static void ccs_collect_entry(void)
 		list_for_each_entry_rcu(ptr, &ccs_policy_manager_list, list) {
 			if (!ptr->is_deleted)
 				continue;
-			if (ccs_add_to_gc(CCS_ID_MANAGER, ptr))
-				list_del_rcu(&ptr->list);
-			else
-				break;
+			if (!ccs_add_to_gc(CCS_ID_MANAGER, &ptr->list))
+				goto unlock;
 		}
 	}
 	{
@@ -552,10 +741,8 @@ static void ccs_collect_entry(void)
 		list_for_each_entry_rcu(ptr, &ccs_aggregator_list, list) {
 			if (!ptr->is_deleted)
 				continue;
-			if (ccs_add_to_gc(CCS_ID_AGGREGATOR, ptr))
-				list_del_rcu(&ptr->list);
-			else
-				break;
+			if (!ccs_add_to_gc(CCS_ID_AGGREGATOR, &ptr->list))
+				goto unlock;
 		}
 	}
 	{
@@ -566,18 +753,14 @@ static void ccs_collect_entry(void)
 						list) {
 				if (!acl->is_deleted)
 					continue;
-				if (ccs_add_to_gc(CCS_ID_ACL, acl))
-					list_del_rcu(&acl->list);
-				else
-					break;
+				if (!ccs_add_to_gc(CCS_ID_ACL, &acl->list))
+					goto unlock;
 			}
 			if (!domain->is_deleted ||
 			    ccs_used_by_task(domain))
 				continue;
-			if (ccs_add_to_gc(CCS_ID_DOMAIN, domain))
-				list_del_rcu(&domain->list);
-			else
-				break;
+			if (!ccs_add_to_gc(CCS_ID_DOMAIN, &domain->list))
+				goto unlock;
 		}
 	}
 	{
@@ -588,19 +771,15 @@ static void ccs_collect_entry(void)
 						list) {
 				if (!member->is_deleted)
 					continue;
-				if (ccs_add_to_gc(CCS_ID_PATH_GROUP_MEMBER,
-						  member))
-					list_del_rcu(&member->list);
-				else
-					break;
+				if (!ccs_add_to_gc(CCS_ID_PATH_GROUP_MEMBER,
+						   &member->list))
+					goto unlock;
 			}
 			if (!list_empty(&group->member_list) ||
 			    atomic_read(&group->users))
 				continue;
-			if (ccs_add_to_gc(CCS_ID_PATH_GROUP, group))
-				list_del_rcu(&group->list);
-			else
-				break;
+			if (!ccs_add_to_gc(CCS_ID_PATH_GROUP, &group->list))
+				goto unlock;
 		}
 	}
 	{
@@ -610,20 +789,16 @@ static void ccs_collect_entry(void)
 			list_for_each_entry_rcu(member, &group->member_list,
 						list) {
 				if (!member->is_deleted)
-					break;
-				if (ccs_add_to_gc(CCS_ID_ADDRESS_GROUP_MEMBER,
-						  member))
-					list_del_rcu(&member->list);
-				else
-					break;
+					continue;
+				if (!ccs_add_to_gc(CCS_ID_ADDRESS_GROUP_MEMBER,
+						   &member->list))
+					goto unlock;
 			}
 			if (!list_empty(&group->member_list) ||
 			    atomic_read(&group->users))
 				continue;
-			if (ccs_add_to_gc(CCS_ID_ADDRESS_GROUP, group))
-				list_del_rcu(&group->list);
-			else
-				break;
+			if (!ccs_add_to_gc(CCS_ID_ADDRESS_GROUP, &group->list))
+				goto unlock;
 		}
 	}
 	{
@@ -634,19 +809,15 @@ static void ccs_collect_entry(void)
 						list) {
 				if (!member->is_deleted)
 					continue;
-				if (ccs_add_to_gc(CCS_ID_NUMBER_GROUP_MEMBER,
-						  member))
-					list_del_rcu(&member->list);
-				else
-					break;
+				if (!ccs_add_to_gc(CCS_ID_NUMBER_GROUP_MEMBER,
+						   &member->list))
+					goto unlock;
 			}
 			if (!list_empty(&group->member_list) ||
 			    atomic_read(&group->users))
 				continue;
-			if (ccs_add_to_gc(CCS_ID_NUMBER_GROUP, group))
-				list_del_rcu(&group->list);
-			else
-				break;
+			if (!ccs_add_to_gc(CCS_ID_NUMBER_GROUP, &group->list))
+				goto unlock;
 		}
 	}
 	{
@@ -654,10 +825,8 @@ static void ccs_collect_entry(void)
 		list_for_each_entry_rcu(ptr, &ccs_reservedport_list, list) {
 			if (!ptr->is_deleted)
 				continue;
-			if (ccs_add_to_gc(CCS_ID_RESERVEDPORT, ptr))
-				list_del_rcu(&ptr->list);
-			else
-				break;
+			if (!ccs_add_to_gc(CCS_ID_RESERVEDPORT, &ptr->list))
+				goto unlock;
 		}
 	}
 	{
@@ -665,10 +834,8 @@ static void ccs_collect_entry(void)
 		list_for_each_entry_rcu(ptr, &ccs_address_list, list) {
 			if (atomic_read(&ptr->users))
 				continue;
-			if (ccs_add_to_gc(CCS_ID_IPV6_ADDRESS, ptr))
-				list_del_rcu(&ptr->list);
-			else
-				break;
+			if (!ccs_add_to_gc(CCS_ID_IPV6_ADDRESS, &ptr->list))
+				goto unlock;
 		}
 	}
 	{
@@ -676,39 +843,36 @@ static void ccs_collect_entry(void)
 		list_for_each_entry_rcu(ptr, &ccs_condition_list, list) {
 			if (atomic_read(&ptr->users))
 				continue;
-			if (ccs_add_to_gc(CCS_ID_CONDITION, ptr))
-				list_del_rcu(&ptr->list);
-			else
-				break;
+			if (!ccs_add_to_gc(CCS_ID_CONDITION, &ptr->list))
+				goto unlock;
 		}
 	}
+	for (i = 0; i < CCS_MAX_HASH; i++) {
+		struct ccs_name_entry *ptr;
+		list_for_each_entry_rcu(ptr, &ccs_name_list[i], list) {
+			if (atomic_read(&ptr->users))
+				continue;
+			if (!ccs_add_to_gc(CCS_ID_NAME, &ptr->list))
+				goto unlock;
+		}
+	}
+ unlock:
+	ccs_read_unlock(idx);
 	mutex_unlock(&ccs_policy_lock);
-	mutex_lock(&ccs_name_list_lock);
-	{
-		int i;
-		for (i = 0; i < CCS_MAX_HASH; i++) {
-			struct ccs_name_entry *ptr;
-			list_for_each_entry_rcu(ptr, &ccs_name_list[i], list) {
-				if (atomic_read(&ptr->users))
-					continue;
-				if (ccs_add_to_gc(CCS_ID_NAME, ptr))
-					list_del_rcu(&ptr->list);
-				else {
-					i = CCS_MAX_HASH;
-					break;
-				}
-			}
-		}
-	}
-	mutex_unlock(&ccs_name_list_lock);
 }
 
-static void ccs_kfree_entry(void)
+/**
+ * ccs_kfree_entry - Delete entries in ccs_gc_list .
+ *
+ * Returns true if some entries were kfree()d, false otherwise.
+ */
+static bool ccs_kfree_entry(void)
 {
 	struct ccs_gc_entry *p;
 	struct ccs_gc_entry *tmp;
 	size_t size = 0;
-	list_for_each_entry_safe(p, tmp, &ccs_gc_queue, list) {
+	bool result = false;
+	list_for_each_entry_safe(p, tmp, &ccs_gc_list, list) {
 		switch (p->type) {
 		case CCS_ID_DOMAIN_INITIALIZER:
 			size = ccs_del_domain_initializer(p->element);
@@ -759,7 +923,10 @@ static void ccs_kfree_entry(void)
 			size = ccs_del_ipv6_address(p->element);
 			break;
 		case CCS_ID_CONDITION:
-			size = ccs_del_condition(p->element);
+			size = ccs_del_condition(container_of(p->element,
+							      struct
+							      ccs_condition,
+							      list));
 			break;
 		case CCS_ID_NAME:
 			size = ccs_del_name(p->element);
@@ -774,37 +941,72 @@ static void ccs_kfree_entry(void)
 			break;
 		default:
 			size = 0;
-			printk(KERN_WARNING "Unknown type\n");
 			break;
 		}
 		ccs_memory_free(p->element, size);
 		list_del(&p->list);
 		kfree(p);
+		ccs_gc_list_len--;
+		result = true;
 	}
+	return result;
 }
 
+/**
+ * ccs_gc_thread - Garbage collector thread function.
+ *
+ * @unused: Unused.
+ *
+ * In case OOM-killer choose this thread for termination, we create this thread
+ * as a short live thread whenever /proc/ccs/ interface was close()d.
+ *
+ * Returns 0.
+ */
 static int ccs_gc_thread(void *unused)
 {
+	static DEFINE_MUTEX(ccs_gc_mutex);
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
 	daemonize("GC for CCS");
 #else
 	daemonize();
 	reparent_to_init();
+#if defined(TASK_DEAD)
+	{
+		struct task_struct *task = current;
+		spin_lock_irq(&task->sighand->siglock);
+		siginitsetinv(&task->blocked, 0);
+		recalc_sigpending();
+		spin_unlock_irq(&task->sighand->siglock);
+	}
+#else
+	{
+		struct task_struct *task = current;
+		spin_lock_irq(&task->sigmask_lock);
+		siginitsetinv(&task->blocked, 0);
+		recalc_sigpending(task);
+		spin_unlock_irq(&task->sigmask_lock);
+	}
+#endif
+	snprintf(current->comm, sizeof(current->comm) - 1, "GC for CCS");
 #endif
 	if (mutex_trylock(&ccs_gc_mutex)) {
-		int i;
-		for (i = 0; i < 10; i++) {
+		do {
 			ccs_collect_entry();
-			if (list_empty(&ccs_gc_queue))
+			if (list_empty(&ccs_gc_list))
 				break;
-			ccs_synchronize_srcu();
-			ccs_kfree_entry();
-		}
+			ccs_synchronize_counter();
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 19)
+			synchronize_srcu(&ccs_ss);
+#endif
+		} while (ccs_kfree_entry());
 		mutex_unlock(&ccs_gc_mutex);
 	}
-	do_exit(0);
+	return 0;
 }
 
+/**
+ * ccs_run_gc - Start garbage collector thread.
+ */
 void ccs_run_gc(void)
 {
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
